
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '35.6.1',
    'branch_nick': 'release-35-LTS',
    'revision_id': 'da061cd8b9e3955a4196b81dc6c0a4b1202bbe20',
    'revno': 13193
}
